@@ -10,31 +10,35 @@ simple_rapply <- function(x, fn, ...) {
 
 `%||%` <- function(x, y) if (is.null(x)) y else x
 
+replace_w_na <- function(x){
+  lapply(x, `%||%`, y = NA)
+}
+
 subs_na <- function(x, type = c("row_df", "col_df", "flat", "rbind_df", "identical"), prefix = NULL) {
   type <- match.arg(type)
-  if (type == "identical"){
-    return(x)
-  }
-
   if (length(x) == 0) {
     return(NA)
   }
 
+  if (type == "identical") {
+    return(x)
+  }
+
   out <- switch(type,
-    row_df = as.data.frame(x),
-    col_df = tibble::enframe(unlist(x)),
+    row_df = as.data.frame(replace_w_na(x)),
     flat = unlist(x),
-    rbind_df = do.call(rbind.data.frame, x)
+    rbind_df = do.call(rbind.data.frame, lapply(x, replace_w_na)
+    )
   )
 
-  if (!is.null(prefix)){
+  if (!is.null(prefix)) {
     out <- prepend(out, prefix)
   }
 
   list(out)
 }
 
-prepend <- function(x, prefix = ""){
+prepend <- function(x, prefix = "") {
   names(x) <- paste(prefix, names(x), sep = "_")
   x
 }
@@ -63,13 +67,16 @@ id_type <- function(identifier) {
     V = "venues",
     I = "institutions",
     C = "concepts",
+    S = "sources",
+    P = "publishers",
+    `F` = "funders",
     NA
   )
 }
 
 oa_email <- function() {
   email <- Sys.getenv("openalexR.mailto")
-  if (email == ""){
+  if (email == "") {
     email <- getOption("openalexR.mailto", default = NULL)
   }
   email
@@ -77,7 +84,7 @@ oa_email <- function() {
 
 oa_apikey <- function() {
   apikey <- Sys.getenv("openalexR.apikey")
-  if (apikey == ""){
+  if (apikey == "") {
     apikey <- getOption("openalexR.apikey", default = NULL)
   }
   apikey
@@ -91,7 +98,9 @@ oa_progress <- function(n, text = "converting") {
 }
 
 asl <- function(z) {
-  if (length(z) > 1) return(z)
+  if (length(z) > 1) {
+    return(z)
+  }
 
   z_low <- tolower(z)
   if (z_low == "true" || z_low == "false") {
@@ -108,3 +117,19 @@ shorten_oaid <- function(id) {
 shorten_orcid <- function(id) {
   gsub("^https://orcid.org/", "", id)
 }
+
+rbind_oa_ls <- function(list_df) {
+  all_names <- unique(unlist(lapply(list_df, names)))
+  do.call(
+    rbind.data.frame,
+    lapply(
+      list_df,
+      function(x) {
+        tibble::as_tibble(c(x, sapply(
+          setdiff(all_names, names(x)),
+          function(y) NA
+        )))
+      }
+    )
+  )
+}
